// Buffered Graphics for C
// Allows for the use of graphics buffers to perform basic operations.
// Works only in Doors mode.

#ifndef __BUF
#define __BUF

#include <doors.h>
#include <all.h>

// Constants --------------------------------------------------------------

#define MaxPlanes 2

#define LCDWidth 240
#define LCDHeight 128
#define LCDByteWidth 30      // = LCDWidth / 8
#define LCDByteSize LCD_SIZE

#define BitOn 1
#define BitOff 0

// Types ------------------------------------------------------------------

// Default types
typedef unsigned char byte;
typedef struct
{
	unsigned char Bit7: 1;
	unsigned char Bit6: 1;
	unsigned char Bit5: 1;
	unsigned char Bit4: 1;
	unsigned char Bit3: 1;
	unsigned char Bit2: 1;
	unsigned char Bit1: 1;
	unsigned char Bit0: 1;
} TBitMaskByte;
typedef TBitMaskByte* PBitMaskByte;

// Graphics buffer for background operations
typedef struct
{
	int PlaneCount;                       // Use InitBuffer to set this
	LCD_BUFFER Planes [MaxPlanes];        // Plane 0 is dark gray plane
} GRAPH_BUFFER;
typedef GRAPH_BUFFER TGraphBuffer;
typedef LCD_BUFFER TLCDBuffer;
typedef TLCDBuffer* PLCDBuffer;

// Common custom graphics buffers
typedef struct
{
	int PlaneCount;                       // Use InitBuffer((TGraphBuffer*)(&Variable),1) to set this
	LCD_BUFFER Plane;
} TBWBuffer;
typedef struct
{
	int PlaneCount;                       // Use InitBuffer((TGraphBuffer*)(&Variable),2) to set this
	LCD_BUFFER Planes [2];                // Plane 0 is dark gray plane
} TGray4Buffer;

// Color definitions
typedef unsigned long TColor;           // Grayscale color using 0x000000 (Black) thru 0xFFFFFF (White)

#define clWhite   0xFFFFFF
#define clSilver  0xAAAAAA
#define clGray    0x555555
#define clBlack   0x000000

#define clGray0   0xFFFFFF
#define clGray17  0xD55555
#define clGray25  0xC00000
#define clGray33  0xAAAAAA
#define clGray50  0x800000
#define clGray67  0x555555
#define clGray75  0x400000
#define clGray83  0x2AAAAA
#define clGray100 0x000000

// Functions --------------------------------------------------------------

void SetCurGrayPlanes(int NewGrayPlanes);
                                        // Specify grayscale mode
                                        // 1: B/W; 2: Gray-4
                                        // Must be set manually
int GetCurGrayPlanes();                 // Return grayscale mode
TLCDBuffer *GetGrayPlane(int Plane);    // Get a plane, or LCD memory in B/W mode
                                        // 0 is dark gray plane

void InitBuf(TGraphBuffer *Buffer, int PlaneCount);
                                        // Initializes a buffer; must be called
                                        // Call after grayscale mode is set
void ClrBuf(TGraphBuffer *Buffer);      // Clears the contents of a buffer
void FlushBuf(TGraphBuffer *Buffer);    // Draws the contents of the buffer on the LCD
void ActivateBuf(TGraphBuffer *Buffer, int Plane);
                                        // Activates a plane in the buffer for standard drawing
                                        // Do not use this unless implementing the function to draw directly into the buffer is impossible or inefficient.
                                        // To to this later, use a function to call this which you can replace later.
void DeactivateBuf();                   // Call after standard drawing is done.

void DrawHLineBuf(TGraphBuffer *Buffer, int X1, int X2, int Y, TColor Color);
                                        // Draws a horizontal line from X1 to X2
void DrawFullHLinesBuf(TGraphBuffer *Buffer, int Y1, int Y2, TColor Color);
                                        // Draws complete horizontal lines from Y1 to Y2, thereby filling the area between Y1 and Y2

int GetBufPos(int X, int Y);
byte GetByteMask(char Start, char End);
void AddByteMask(byte *Ptr, byte Mask);
void ClrByteMask(byte *Ptr, byte Mask);

// Variables --------------------------------------------------------------

int CurGrayPlanes = 1;

// Function Implementations -----------------------------------------------

void SetCurGrayPlanes(int NewGrayPlanes)
{
	CurGrayPlanes = NewGrayPlanes;
}

int GetCurGrayPlanes()
{
	return (CurGrayPlanes);
}

TLCDBuffer *GetGrayPlane(int Plane)
{
	if (CurGrayPlanes <= 1)
		return LCD_MEM;
	else
		return GetPlane (CurGrayPlanes - Plane - 1);
}

void InitBuf(TGraphBuffer *Buffer, int PlaneCount)
{
	Buffer->PlaneCount = PlaneCount;
	ClrBuf (Buffer);
}

void ClrBuf(TGraphBuffer *Buffer)
{
	int I;
	for (I = 0; I < Buffer->PlaneCount; I++)
		memset (&(Buffer->Planes [I]), 0, LCDByteSize);
}

void FlushBuf(TGraphBuffer *Buffer)
{
	int I;
	for (I = 0; I < Buffer->PlaneCount; I++)
		memcpy (GetGrayPlane (I), &(Buffer->Planes [I]), LCDByteSize);
}

void ActivateBuf(TGraphBuffer *Buffer, int Plane)
{
	PortSet (&(Buffer->Planes [Plane]), LCDWidth - 1, LCDHeight - 1);
}

void DeactivateBuf()
{
	PortRestore();
}

void DrawHLineBuf(TGraphBuffer *Buffer, int X1, int X2, int Y, TColor Color)
{
	// I know this all looks awful.  So if you have got something better and more efficient, please let me know.
	int X1Pos, X2Pos;
	byte MaskByte1, MaskByte2;
	if (Y < 0) Y = 0;
	if (Y >= LCDHeight) Y = LCDHeight - 1;
	if (X1 < 0) X1 = 0;
	if (X2 >= LCDWidth) X2 = LCDWidth - 1;
	
	if (X1 <= X2)
	{
		X1Pos = GetBufPos (X1, Y) + 1;
		X2Pos = GetBufPos (X2, Y) + 1;
		MaskByte1 = GetByteMask (0, 7 - (X1 % 8));
		MaskByte2 = GetByteMask (7 - (X2 % 8), 7);
		if (Color < clGray50)      // "<" means "darker than"
		{
			// Draw it in the dark gray plane
			AddByteMask (&(Buffer->Planes [0] [X1Pos - 1]), MaskByte1);
			memset (&(Buffer->Planes [0] [X1Pos]), 0xFF, X2Pos - X1Pos);
			AddByteMask (&(Buffer->Planes [0] [X2Pos]), MaskByte2);
			if (Buffer->PlaneCount > 1)
			{
				if (Color < clGray83)  // clGray100 is darkest
				{
					// Draw it in the light gray plane
					AddByteMask (&(Buffer->Planes [1] [X1Pos - 1]), MaskByte1);
					memset (&(Buffer->Planes [1] [X1Pos]), 0xFF, X2Pos - X1Pos);
					AddByteMask (&(Buffer->Planes [1] [X2Pos]), MaskByte2);
				}
				else
				{
					// Erase it in the light gray plane
					ClrByteMask (&(Buffer->Planes [1] [X1Pos - 1]), MaskByte1);
					memset (&(Buffer->Planes [1] [X1Pos]), 0x00, X2Pos - X1Pos);
					ClrByteMask (&(Buffer->Planes [1] [X2Pos]), MaskByte2);
				}
			}
		}
		else
		{
			// Erase it in the dark gray plane
			ClrByteMask (&(Buffer->Planes [0] [X1Pos - 1]), MaskByte1);
			memset (&(Buffer->Planes [0] [X1Pos]), 0x00, X2Pos - X1Pos);
			ClrByteMask (&(Buffer->Planes [0] [X2Pos]), MaskByte2);
			if (Buffer->PlaneCount > 1)
			{
				if (Color < clGray17)
				{
					// Draw it in the light gray plane
					AddByteMask (&(Buffer->Planes [1] [X1Pos - 1]), MaskByte1);
					memset (&(Buffer->Planes [1] [X1Pos]), 0xFF, X2Pos - X1Pos);
					AddByteMask (&(Buffer->Planes [1] [X2Pos]), MaskByte2);
				}
				else
				{
					// Erase it in the light gray plane
					ClrByteMask (&(Buffer->Planes [1] [X1Pos - 1]), MaskByte1);
					memset (&(Buffer->Planes [1] [X1Pos]), 0x00, X2Pos - X1Pos);
					ClrByteMask (&(Buffer->Planes [1] [X2Pos]), MaskByte2);
				}
			}
		}
	}
}

void DrawFullHLinesBuf(TGraphBuffer *Buffer, int Y1, int Y2, TColor Color)
{
	int Pos, Len, Tmp;
	if (Y1 > Y2)
	{
		Tmp = Y1;
		Y1 = Y2;
		Y2 = Tmp;
	}
	Y2++;
	if (Y1 < 0) Y1 = 0;
	if (Y2 > LCDHeight) Y2 = LCDHeight;
	Pos = GetBufPos (0, Y1);
	Len = GetBufPos (0, Y2) - Pos;
	
	if (Color < clGray50)      // "<" means "darker than"
	{
		// Draw it in the dark gray plane
		memset (&(Buffer->Planes [0] [Pos]), 0xFF, Len);
		if (Buffer->PlaneCount > 1)
		{
			if (Color < clGray83)  // clGray100 is darkest
			{
				// Draw it in the light gray plane
				memset (&(Buffer->Planes [1] [Pos]), 0xFF, Len);
			}
			else
			{
				// Erase it in the light gray plane
				memset (&(Buffer->Planes [1] [Pos]), 0x00, Len);
			}
		}
	}
	else
	{
		// Erase it in the dark gray plane
		memset (&(Buffer->Planes [0] [Pos]), 0x00, Len);
		if (Buffer->PlaneCount > 1)
		{
			if (Color < clGray17)
			{
				// Draw it in the light gray plane
				memset (&(Buffer->Planes [1] [Pos]), 0xFF, Len);
			}
			else
			{
				// Erase it in the light gray plane
				memset (&(Buffer->Planes [1] [Pos]), 0x00, Len);
			}
		}
	}
}

int GetBufPos(int X, int Y)
{
	return (Y * LCDByteWidth + X / 8);
}

byte GetByteMask(char Start, char End)
{
	TBitMaskByte Res;
	memset (&Res, 0x00, 1);
	if ((Start <= 0) && (End >= 0))
		Res.Bit0 = BitOn;
	if ((Start <= 1) && (End >= 1))
		Res.Bit1 = BitOn;
	if ((Start <= 2) && (End >= 2))
		Res.Bit2 = BitOn;
	if ((Start <= 3) && (End >= 3))
		Res.Bit3 = BitOn;
	if ((Start <= 4) && (End >= 4))
		Res.Bit4 = BitOn;
	if ((Start <= 5) && (End >= 5))
		Res.Bit5 = BitOn;
	if ((Start <= 6) && (End >= 6))
		Res.Bit6 = BitOn;
	if ((Start <= 7) && (End >= 7))
		Res.Bit7 = BitOn;
	return *((byte*)(&Res));
}

void AddByteMask(byte *Ptr, byte Mask)
{
	*Ptr = *Ptr | Mask;
}

void ClrByteMask(byte *Ptr, byte Mask)
{
	*Ptr = *Ptr & (0xFF - Mask);
}

#endif
